;;; get-opt.scm - Argument option processor
;;;
;;; Copyright (c) 2002 Anthony Carrico
;;;
;;; All rights reserved.
;;;
;;; Redistribution and use in source and binary forms, with or without
;;; modification, are permitted provided that the following conditions
;;; are met:
;;; 1. Redistributions of source code must retain the above copyright
;;;    notice, this list of conditions and the following disclaimer.
;;; 2. Redistributions in binary form must reproduce the above copyright
;;;    notice, this list of conditions and the following disclaimer in the
;;;    documentation and/or other materials provided with the distribution.
;;; 3. The name of the authors may not be used to endorse or promote products
;;;    derived from this software without specific prior written permission.
;;;
;;; THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
;;; IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
;;; OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
;;; IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
;;; INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
;;; NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
;;; DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
;;; THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
;;; (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
;;; THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

(define-record-type option :option
  ;; option processor signature:
  (option
   names         ; list of long option strings and short option chars
   required-arg?
   optional-arg?
   processor)    ; (lambda (option name arg . seeds) ...) -> next-seed ...
  option?
  (names option-names)
  (required-arg? option-required-arg?)
  (optional-arg? option-optional-arg?)
  (processor option-processor))

(define args-fold
  (lambda (args              ; list of args
           options           ; list of options
           unrecognized-option-proc
           ;; (lambda (non-option . seeds) <body>) -> next-seed ...
           non-option-proc
           . seeds)
    (letrec
        ((find
          (lambda (l ?)
            (cond ((null? l) #f)
                  ((? (car l)) (car l))
                  (else (find (cdr l) ?)))))
         (find-option
          ;; ISSUE: This is a brute force search. Could use a table.
          (lambda (name)
            (find
             options
             (lambda (option)
               (find
                (option-names option)
                (lambda (test-name)
                  (equal? name test-name)))))))
         (scan-short-options
          (lambda (index shorts args seeds)
            (if (= index (string-length shorts))
                (scan-args args seeds)
                (let* ((name (string-ref shorts index))
                       (option (or (find-option name)
                                   (option (list name)
                                           #f
                                           #f
                                           unrecognized-option-proc))))
                  (cond ((and (< (+ index 1) (string-length shorts))
                              (or (option-required-arg? option)
                                  (option-optional-arg? option)))
                         (let-values
                             ((seeds (apply (option-processor option)
                                            option
                                            name
                                            (substring
                                             shorts
                                             (+ index 1)
                                             (string-length shorts))
                                            seeds)))
                           (scan-args args seeds)))
                        ((and (option-required-arg? option)
                              (pair? args))
                         (let-values
                             ((seeds (apply (option-processor option)
                                            option
                                            name
                                            (car args)
                                            seeds)))
                           (scan-args (cdr args) seeds)))
                        (else
                         (let-values
                             ((seeds (apply (option-processor option)
                                            option
                                            name
                                            #f
                                            seeds)))
                           (scan-short-options
                            (+ index 1)
                            shorts
                            args
                            seeds))))))))
	 (scan-non-options
	  (lambda (non-options seeds)
	    (if (null? non-options)
		(apply values seeds)
		(let-values ((seeds (apply non-option-proc (car non-options) seeds)))
		  (scan-non-options (cdr non-options) seeds)))))
         (scan-args
          (lambda (args seeds)
            (if (null? args)
                (apply values seeds)
                (let ((arg (car args))
                      (args (cdr args)))
                  (match-cond
                   ((regexp-search (rx bos "--" eos) arg)
                    (whole)
                    ;; End option scanning:
		    (scan-non-options args seeds))
                   ((regexp-search (rx bos
                                       "--"
                                       (submatch (+ (~ "=")))
                                       "="
                                       (submatch (* any)))
                                   arg)
                    (whole name arg)
                    ;; Found long option with arg:
                    (let*-values
                     (((option) (or (find-option name)
                                    (option (list name)
                                            #t
                                            #f
                                            unrecognized-option-proc)))
                      (seeds (apply (option-processor option)
                                    option
                                    name
                                    arg
                                    seeds)))
                     (scan-args args seeds)))
                   ((regexp-search (rx bos "--" (submatch (+ any))) arg)
                    (whole name)
                    ;; Found long option:
                    (let ((option (or (find-option name)
                                      (option
                                       (list name)
                                       #f
                                       #f
                                       unrecognized-option-proc))))
                      (if (and (option-required-arg? option)
                               (pair? args))
                          (let-values
                              ((seeds (apply (option-processor option)
                                             option
                                             name
                                             (car args)
                                             seeds)))
                            (scan-args (cdr args) seeds))
                          (let-values
                              ((seeds (apply (option-processor option)
                                             option
                                             name
                                             #f
                                             seeds)))
                            (scan-args args seeds)))))
                   ((regexp-search (rx bos "-" (submatch (+ any))) arg)
                    (whole shorts)
                    ;; Found short options
                    (scan-short-options 0 shorts args seeds))
                   (else
		    (let-values ((seeds (apply non-option-proc arg seeds)))
		      (scan-args args seeds)))))))))
      (scan-args args seeds))))
